class SeldonModel:
    """
    Model template. You can load your model parameters in __init__ from a location
    accessible at runtime.
    """

    def __init__(self):
        """
        Add any initialization parameters. These will be passed at runtime from
        the graph definition parameters defined in your seldondeployment kubernetes
        resource manifest.
        """
        print("Initializing")

    def predict(self, X, features_names):
        """
        Return a prediction.

        Parameters
        ----------
        X : array-like
        feature_names : array of feature names (optional)
        """
        print("Predict called - will run idenity function")
        return X

    def send_feedback(self, features, feature_names, reward, truth):
        """
        Handle feedback

        Parameters
        ----------
        features : array - the features sent in the original predict request
        feature_names : array of feature names. May be None if not available.
        reward : float - the reward
        truth : array with correct value (optional)
        """
        print("Send feedback called")
        return []


seldon_model = SeldonModel



import asyncio
import logging
import sys
from typing import AsyncIterator, Optional

import neuro_sdk as neuro_api


if sys.version_info >= (3, 7):  # pragma: no cover
    from contextlib import asynccontextmanager
else:
    from async_generator import asynccontextmanager

logger = logging.getLogger(__name__)


@asynccontextmanager
async def get_neuro_client(
    cluster: Optional[str] = None,
) -> AsyncIterator[neuro_api.Client]:
    client: neuro_api.Client = await neuro_api.get()
    cluster_orig: Optional[str] = None
    try:
        await client.__aenter__()

        cluster_orig = client.cluster_name
        if cluster is not None:
            if cluster != cluster_orig:
                logger.info(
                    f"Temporarily switching cluster: {cluster_orig} -> {cluster}"
                )
                await client.config.switch_cluster(cluster)  # typing: ignore
            else:
                logger.info(f"Already on cluster: {cluster}")
        yield client
    finally:
        # NOTE: bypass https://github.com/neuro-inc/platform-client-python/issues/1816
        try:
            await client.__aexit__(None, None, None)
        except asyncio.CancelledError:
            raise
        except BaseException as e:
            logger.warning(f"Ignoring exception during closing neuro client: {e}")

        if cluster is not None and cluster_orig is not None and cluster != cluster_orig:
            logger.info(f"Switching back cluster: {cluster} -> {cluster_orig}")
            try:
                await client.config.switch_cluster(cluster_orig)
            except asyncio.CancelledError:
                raise
            except BaseException:
                logger.error(
                    f"Could not switch back to cluster '{cluster_orig}'. Please "
                    f"run manually: 'neuro config switch-cluster {cluster_orig}'"
                )



import logging
import sys
import tempfile
import textwrap
from pathlib import Path
from typing import Optional, Sequence, Tuple

import click
import neuro_sdk as neuro_api
from neuro_cli.asyncio_utils import run as run_async
from neuro_cli.const import EX_OK, EX_PLATFORMERROR
from neuro_sdk.client import Client
from neuro_sdk.url_utils import uri_from_cli

from .cli import main
from .image_builder import DockerConfigAuth, ImageBuilder, create_docker_config_auth
from .utils import get_neuro_client


logger = logging.getLogger(__name__)


@main.group()
def image() -> None:
    """
    Job container image operations.
    """
    pass


@image.command("transfer")
@click.argument("source")
@click.argument("destination")
@click.option(
    "-F",
    "--force-overwrite",
    default=False,
    is_flag=True,
    help="Transfer even if the destination image already exists.",
)
def image_transfer(source: str, destination: str, force_overwrite: bool) -> None:
    """
    Copy images between clusters.
    """
    exit_code = run_async(_image_transfer(source, destination, force_overwrite))
    sys.exit(exit_code)


@image.command(
    "build", help="Build Job container image remotely on cluster using Kaniko."
)
@click.argument("path", metavar="CONTEXT_PATH")
@click.argument("image_uri")
@click.option(
    "-f",
    "--file",
    default="Dockerfile",
    show_default=True,
    help=(
        "Relative (w.r.t. context) path to the dockerfile. "
        "The dockerfile should be within the context directory."
    ),
)
@click.option(
    "--build-arg",
    multiple=True,
    metavar="VAR=VAL",
    help=(
        "Buid-time variables passed in ARG values, similarly to Docker. "
        "Could be used multiple times for multiple arguments."
    ),
)
@click.option(
    "-v",
    "--volume",
    metavar="MOUNT",
    multiple=True,
    help=(
        "Mounts directory from vault into container. "
        "Use multiple options to mount more than one volume. "
        "Use --volume=ALL to mount all accessible storage directories."
    ),
)
@click.option(
    "-e",
    "--env",
    metavar="VAR=VAL",
    multiple=True,
    help=(
        "Set environment variable in container "
        "Use multiple options to define more than one variable"
    ),
)
@click.option(
    "-s",
    "--preset",
    metavar="PRESET",
    help=(
        "Predefined resource configuration (to see available values, "
        "run `neuro config show`)"
    ),
)
@click.option(
    "-F",
    "--force-overwrite",
    default=False,
    show_default=True,
    is_flag=True,
    help="Overwrite if the destination image already exists.",
)
@click.option(
    "--cache/--no-cache",
    default=True,
    show_default=True,
    help="Use Kaniko cache while building image.",
)
@click.option(
    "--verbose",
    type=bool,
    default=False,
    help="If specified, run Kaniko with 'debug' verbosity, otherwise 'info' (default).",
)
@click.option(
    "--build-tag",
    multiple=True,
    metavar="VAR=VAL",
    help=(
        "Set tag(s) for image builder job. "
        "We will add tag 'kaniko-builds:{image-name}' authomatically."
    ),
)
def image_build(
    path: str,
    image_uri: str,
    file: str,
    build_arg: Tuple[str],
    volume: Tuple[str],
    env: Tuple[str],
    preset: str,
    force_overwrite: bool,
    cache: bool,
    verbose: bool,
    build_tag: Tuple[str],
) -> None:
    try:
        sys.exit(
            run_async(
                _build_image(
                    dockerfile_path=Path(file),
                    context=path,
                    image_uri_str=image_uri,
                    use_cache=cache,
                    build_args=build_arg,
                    volume=volume,
                    env=env,
                    preset=preset,
                    force_overwrite=force_overwrite,
                    verbose=verbose,
                    build_tags=build_tag,
                )
            )
        )
    except (ValueError, click.ClickException) as e:
        logger.error(f"Failed to build image: {e}")
        sys.exit(EX_PLATFORMERROR)


async def _parse_neuro_image(image: str) -> neuro_api.RemoteImage:
    async with get_neuro_client() as client:
        return client.parse.remote_image(image)


def _get_cluster_from_uri(image_uri: str, *, scheme: str) -> Optional[str]:
    try:
        uri = uri_from_cli(image_uri, "", "", allowed_schemes=[scheme])
        return uri.host
    except ValueError:
        # seems like the image scheme was not provided, since it's hosted in dockerhub
        logger.warning(f"Unable to parse the cluster name from URI '{image_uri}' ")
        return None


async def _image_transfer(
    src_uri_str: str, dst_uri_str: str, force_overwrite: bool
) -> int:
    src_cluster: Optional[str] = _get_cluster_from_uri(src_uri_str, scheme="image")
    dst_cluster: Optional[str] = _get_cluster_from_uri(dst_uri_str, scheme="image")
    if not dst_cluster:
        raise ValueError(
            f"Invalid destination image {dst_uri_str}: missing cluster name"
        )

    with tempfile.TemporaryDirectory() as tmpdir:
        async with get_neuro_client(cluster=src_cluster) as src_client:
            src_image = src_client.parse.remote_image(image=src_uri_str)
            src_reg_auth = await create_docker_config_auth(src_client.config)

        dockerfile = Path(f"{tmpdir}/Dockerfile")
        dockerfile.write_text(
            textwrap.dedent(
                f"""\
                FROM {src_image.as_docker_url()}
                LABEL neu.ro/source-image-uri={src_uri_str}
                """
            )
        )
        migration_job_tags = (
            f"src-image:{src_image}",
            f"neuro-extras:image-transfer",
        )
        return await _build_image(
            dockerfile_path=Path(dockerfile.name),
            context=tmpdir,
            image_uri_str=dst_uri_str,
            use_cache=True,
            build_args=(),
            volume=(),
            env=(),
            build_tags=migration_job_tags,
            force_overwrite=force_overwrite,
            registry_auths=[src_reg_auth],
        )


async def _build_image(
    dockerfile_path: Path,
    context: str,
    image_uri_str: str,
    use_cache: bool,
    build_args: Tuple[str, ...],
    volume: Tuple[str, ...],
    env: Tuple[str, ...],
    build_tags: Tuple[str, ...],
    force_overwrite: bool,
    preset: Optional[str] = None,
    registry_auths: Sequence[DockerConfigAuth] = (),
    verbose: bool = False,
) -> int:
    cluster = _get_cluster_from_uri(image_uri_str, scheme="image")
    async with get_neuro_client(cluster=cluster) as client:
        context_uri = uri_from_cli(
            context,
            client.username,
            client.cluster_name,
            allowed_schemes=("file", "storage"),
        )
        image_exists = await _check_image_exists(image_uri_str, client)
        if image_exists and force_overwrite:
            logger.warning(
                f"Target image '{image_uri_str}' exists and will be overwritten."
            )
        elif image_exists and not force_overwrite:
            raise click.ClickException(
                f"Target image '{image_uri_str}' exists. "
                f"Use -F/--force-overwrite flag to enforce overwriting."
            )

        builder = ImageBuilder(
            client, extra_registry_auths=registry_auths, verbose=verbose
        )
        exit_code = await builder.build(
            dockerfile_path=dockerfile_path,
            context_uri=context_uri,
            image_uri_str=image_uri_str,
            use_cache=use_cache,
            build_args=build_args,
            volumes=volume,
            envs=env,
            job_preset=preset,
            build_tags=build_tags,
        )
        if exit_code == EX_OK:
            logger.info(f"Successfully built {image_uri_str}")
            return EX_OK
        else:
            raise click.ClickException(f"Failed to build image: {exit_code}")


async def _check_image_exists(image_uri_str: str, client: Client) -> bool:
    image = await _parse_neuro_image(image_uri_str)
    if image.registry is None:
        # TODO (y.s.): we might need to implement this check later.
        logger.warning(
            f"Skipping check if image '{image}' exists. "
            "If it does exist - it will be overwritten!"
        )
        return False
    try:
        existing_images = await client.images.tags(
            neuro_api.RemoteImage(
                name=image.name,
                owner=image.owner,
                cluster_name=image.cluster_name,
                registry=image.registry,
                tag=None,
            )
        )
        return image in existing_images
    except neuro_api.errors.ResourceNotFound:
        # image does not exists on platform registry
        return False



import json
from typing import Optional

import click
from neuro_cli.asyncio_utils import run as run_async
from neuro_sdk.url_utils import uri_from_cli

from .cli import main
from .image_builder import DockerConfigAuth, ImageBuilder
from .utils import get_neuro_client


@main.group()
def config() -> None:
    """
    Configuration operations.
    """
    pass


@config.command(
    "save-registry-auth",
    help="Save docker auth file for accessing platform registry.",
)
@click.option(
    "--cluster",
    help=(
        "Cluster name for which the auth information should be saved. "
        "Current cluster by default"
    ),
)
@click.argument("path")
def save_registry_auth(path: str, cluster: Optional[str]) -> None:
    run_async(_save_registry_auth(path, cluster))


async def _save_registry_auth(path: str, cluster: Optional[str]) -> None:
    async with get_neuro_client(cluster) as client:
        uri = uri_from_cli(
            path,
            client.username,
            client.cluster_name,
            allowed_schemes=("file", "storage"),
        )
        builder = ImageBuilder(client)
        docker_config = await builder.create_docker_config()
        click.echo(f"Saving Docker config.json as {uri}")
        if uri.scheme == "file":
            with open(path, "w") as f:
                json.dump(docker_config.to_primitive(), f)
        else:
            await builder.save_docker_config(docker_config, uri)


@config.command("build-registy-auth")
@click.argument("registry-uri")
@click.argument("username")
@click.argument("password")
def build_registy_auth(registry_uri: str, username: str, password: str) -> None:
    """Generate docker auth for accessing remote registry."""
    auth = _build_registy_auth(registry_uri, username, password)
    click.echo(auth)


def _build_registy_auth(registry_uri: str, username: str, password: str) -> str:
    config = DockerConfigAuth(registry_uri, username, password)
    result = {"auths": {registry_uri: {"auth": config.credentials}}}
    return json.dumps(result)



import asyncio
from pathlib import Path

import click
from neuro_cli.asyncio_utils import run as run_async
from neuro_sdk import ConfigError, Factory, find_project_root

from .cli import main


@main.command("upload")
@click.argument("path")
def upload(path: str) -> None:
    """
    Upload neuro project files to storage.

    Uploads file (or files under) project-root/PATH to
    storage://remote-project-dir/PATH. You can use "." for PATH to upload
    whole project. The "remote-project-dir" is set using .neuro.toml config,
    as in example:

    \b
    [extra]
    remote-project-dir = "project-dir-name"
    """
    return_code = run_async(_upload(path))
    exit(return_code)


@main.command("download")
@click.argument("path")
def download(path: str) -> None:
    """
    Download neuro project files from storage.

    Downloads file (or files under) from storage://remote-project-dir/PATH
    to project-root/PATH. You can use "." for PATH to download whole project.
    The "remote-project-dir" is set using .neuro.toml config, as in example:

    \b
    [extra]
    remote-project-dir = "project-dir-name"
    """
    return_code = run_async(_download(path))
    exit(return_code)


async def _upload(path: str) -> int:
    target = _get_project_root() / path
    if not target.exists():
        raise click.ClickException(f"Folder or file does not exist: {target}")
    remote_project_root = await _get_remote_project_root()
    await _ensure_folder_exists((remote_project_root / path).parent, True)
    if target.is_dir():
        subprocess = await asyncio.create_subprocess_exec(
            "neuro",
            "cp",
            "--recursive",
            "-u",
            str(target),
            "-T",
            f"storage:{remote_project_root / path}",
        )
    else:
        subprocess = await asyncio.create_subprocess_exec(
            "neuro", "cp", str(target), f"storage:{remote_project_root / path}"
        )
    return await subprocess.wait()


async def _download(path: str) -> int:
    project_root = _get_project_root()
    remote_project_root = await _get_remote_project_root()
    await _ensure_folder_exists((project_root / path).parent, False)
    subprocess = await asyncio.create_subprocess_exec(
        "neuro",
        "cp",
        "--recursive",
        "-u",
        f"storage:{remote_project_root / path}",
        "-T",
        str(project_root / path),
    )
    return await subprocess.wait()


def _get_project_root() -> Path:
    try:
        return find_project_root()
    except ConfigError:
        raise click.ClickException(
            "Not a Neu.ro project directory (or any of the parent directories)."
        )


async def _get_remote_project_root() -> Path:
    factory = Factory()
    config = await factory.load_user_config()
    try:
        return Path(config["extra"]["remote-project-dir"])
    except KeyError:
        raise click.ClickException(
            '"remote-project-dir" configuration variable is not set. Please add'
            ' it to "extra" section of project config file.'
        )


async def _ensure_folder_exists(path: Path, remote: bool = False) -> None:
    if remote:
        subprocess = await asyncio.create_subprocess_exec(
            "neuro", "mkdir", "-p", f"storage:{path}"
        )
        returncode = await subprocess.wait()
        if returncode != 0:
            raise click.ClickException("Was unable to create containing directory")
    else:
        path.mkdir(parents=True, exist_ok=True)



import logging

import click
from neuro_cli.asyncio_utils import setup_child_watcher

from .version import __version__


setup_child_watcher()


class ClickLogHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        try:
            msg = self.format(record)
            click.echo(msg)
        except Exception:
            self.handleError(record)


@click.group()
@click.version_option(
    version=__version__, message="neuro-extras package version: %(version)s"
)
def main() -> None:
    """
    Auxiliary scripts and recipes for automating routine tasks.
    """
    handler = ClickLogHandler()
    handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))

    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)
    root_logger.addHandler(handler)



import asyncio
import logging
import os
import tempfile
from distutils import dir_util
from enum import Enum
from pathlib import Path
from typing import List, Optional, Sequence
from urllib import parse

import click
import neuro_sdk as neuro_api
from neuro_cli.asyncio_utils import run as run_async
from neuro_cli.const import EX_OK
from yarl import URL

from .cli import main
from .common import NEURO_EXTRAS_IMAGE, _attach_job_stdout
from .image import _get_cluster_from_uri, _parse_neuro_image
from .utils import get_neuro_client


SUPPORTED_ARCHIVE_TYPES = (
    ".tar.gz",
    ".tgz",
    ".tar.bz2",
    ".tbz2",
    ".tbz",
    ".tar",
    ".gz",
    ".zip",
)
SUPPORTED_OBJECT_STORAGE_SCHEMES = {
    "AWS": "s3://",
    "GCS": "gs://",
    # originally, Azure's blob scheme is 'http(s)',
    # but we prepend 'azure+' to differenciate https vs azure
    "AZURE": "azure+https://",
    "HTTP": "http://",
    "HTTPS": "https://",
}

TEMP_UNPACK_DIR = Path.home() / ".neuro-tmp"

logger = logging.getLogger(__name__)


class DataCopier:
    def __init__(self, client: neuro_api.Client):
        self._client = client

    async def launch(
        self,
        extract: bool,
        src_uri: URL,
        dst_uri: URL,
        volume: Sequence[str],
        env: Sequence[str],
        use_temp_dir: bool,
        preset: Optional[str] = None,
        life_span: Optional[int] = None,
    ) -> neuro_api.JobDescription:
        logger.info("Submitting a copy job")

        image = await _parse_neuro_image(NEURO_EXTRAS_IMAGE)

        command = self._build_command(dst_uri, src_uri, extract, use_temp_dir)

        env_parse_result = self._client.parse.envs(env)

        vol = self._client.parse.volumes(volume)
        volumes, secret_files, disk_volumes = (
            list(vol.volumes),
            list(vol.secret_files),
            list(vol.disk_volumes),
        )

        job = await self._client.jobs.start(
            image=image,
            preset_name=preset or list(self._client.presets.keys())[0],
            command=command,
            env=env_parse_result.env,
            secret_env=env_parse_result.secret_env,
            volumes=volumes,
            secret_files=secret_files,
            disk_volumes=disk_volumes,
            life_span=life_span or 60 * 60,
        )

        logger.info(f"The copy job ID: {job.id}")
        return job

    def _build_command(
        self, dst_uri: URL, src_uri: URL, extract: bool, use_temp_dir: bool
    ) -> str:
        args = f"{str(src_uri)} {str(dst_uri)}"
        if extract:
            args = f"-x {args}"
        if use_temp_dir:
            args = f"--use-temp-dir {args}"
        return f"neuro-extras data cp {args}"


class UrlType(Enum):
    UNSUPPORTED = 0
    LOCAL = 1
    CLOUD = 2
    STORAGE = 3
    DISK = 4

    @staticmethod
    def get_type(url: URL) -> "UrlType":
        if url.scheme == "storage":
            return UrlType.STORAGE
        if url.scheme == "":
            return UrlType.LOCAL
        if url.scheme in ("s3", "gs", "azure+https", "http", "https"):
            return UrlType.CLOUD
        if url.scheme == "disk":
            return UrlType.DISK
        return UrlType.UNSUPPORTED


@main.group()
def data() -> None:
    """
    Data transfer operations.
    """
    pass


@data.command("transfer")
@click.argument("source")
@click.argument("destination")
def data_transfer(source: str, destination: str) -> None:
    """
    Copy data between storages on different clusters. \n
    Consider archiving dataset first for the sake of performance,
    if the dataset contains a lot (100k+) of small (< 100Kb each) files.
    """
    run_async(_data_transfer(source, destination))


@data.command(
    "cp",
    help=(
        "Copy data between external object storage and cluster. "
        "Supported external object storage systems: "
        f"{list(SUPPORTED_OBJECT_STORAGE_SCHEMES.keys())}. "
        "Note: originally, Azure's blob storage scheme is 'http(s)', "
        "but we prepend 'azure+' to differenciate https vs azure"
    ),
)
@click.argument("source")
@click.argument("destination")
@click.option(
    "-x",
    "--extract",
    default=False,
    is_flag=True,
    help=(
        "Perform extraction of SOURCE into the DESTINATION directory. "
        "The archive type is derived from the file name. "
        f"Supported types: {', '.join(SUPPORTED_ARCHIVE_TYPES)}."
    ),
)
@click.option(
    "-c",
    "--compress",
    default=False,
    is_flag=True,
    help=(
        "Perform compression of SOURCE into the DESTINATION file. "
        "The archive type is derived from the file name. "
        f"Supported types: {', '.join(SUPPORTED_ARCHIVE_TYPES)}."
    ),
)
@click.option(
    "-v",
    "--volume",
    metavar="MOUNT",
    multiple=True,
    help=(
        "Mounts directory from vault into container. "
        "Use multiple options to mount more than one volume."
    ),
)
@click.option(
    "-e",
    "--env",
    metavar="VAR=VAL",
    multiple=True,
    help=(
        "Set environment variable in container. "
        "Use multiple options to define more than one variable."
    ),
)
@click.option(
    "-t",
    "--use-temp-dir",
    default=False,
    is_flag=True,
    help=(
        "Download and extract / compress data (if needed) "
        " inside the temporal directory. "
        "Afterwards move resulted file(s) into the DESTINATION. "
        "NOTE: use it if 'storage:' is involved and "
        "extraction or compression is performed to speedup the process."
    ),
)
@click.option(
    "-s",
    "--preset",
    metavar="PRESET_NAME",
    help=("Preset name used for copy."),
)
@click.option(
    "-l",
    "--life_span",
    metavar="SECONDS",
    help=("Copy job life span in seconds."),
)
def data_cp(
    source: str,
    destination: str,
    extract: bool,
    compress: bool,
    volume: Sequence[str],
    env: Sequence[str],
    use_temp_dir: bool,
    preset: Optional[str] = None,
    life_span: Optional[int] = None,
) -> None:
    if extract and compress:
        raise click.ClickException("Extract and compress can't be used together")
    run_async(
        _data_cp(
            source,
            destination,
            extract,
            compress,
            list(volume),
            list(env),
            use_temp_dir,
            preset,
            life_span,
        )
    )


async def _data_cp(
    source: str,
    destination: str,
    extract: bool,
    compress: bool,
    volume: List[str],
    env: List[str],
    use_temp_dir: bool,
    preset: Optional[str] = None,
    life_span: Optional[int] = None,
) -> None:
    source_url = URL(source)
    destination_url = URL(destination)
    source_url_type = UrlType.get_type(source_url)
    if source_url_type == UrlType.UNSUPPORTED:
        raise ValueError(f"Unsupported source URL scheme: {source_url.scheme}")
    destination_url_type = UrlType.get_type(destination_url)
    if destination_url_type == UrlType.UNSUPPORTED:
        raise ValueError(
            f"Unsupported destination URL scheme: {destination_url.scheme}"
        )

    if source_url_type == UrlType.CLOUD and destination_url_type == UrlType.CLOUD:
        raise ValueError(
            "This command can't be used to copy data between cloud providers"
        )
    if source_url_type == UrlType.STORAGE and destination_url_type == UrlType.STORAGE:
        raise ValueError(
            "This command can't be used to copy data between two storage locations"
        )
    if source_url_type == UrlType.DISK and destination_url_type == UrlType.DISK:
        raise ValueError(
            "This command can't be used to copy data between two persistent disks"
        )
    if destination_url.scheme in ("http", "https"):
        raise ValueError("This command can't be used to upload data over HTTP(S)")

    # Persistent disk and storage locations must be mounted as folders to a job
    if UrlType.STORAGE in (source_url_type, destination_url_type) or UrlType.DISK in (
        source_url_type,
        destination_url_type,
    ):
        if source_url_type == UrlType.STORAGE:
            volume.append(f"{str(source_url)}:/var/storage")
            container_src_uri = URL("/var/storage")
        elif source_url_type == UrlType.DISK:
            volume.append(f"{str(source_url)}:/var/disk:rw")
            container_src_uri = URL("/var/disk/")
        else:
            container_src_uri = source_url

        if destination_url_type == UrlType.STORAGE:
            volume.append(f"{str(destination_url.parent)}:/var/storage")
            container_dst_uri = URL(f"/var/storage/{destination_url.name}")
        elif destination_url_type == UrlType.DISK:
            disk_id = str(destination_url.path)[:41]  # disk ID is 41 symbols long
            volume.append(f"disk:{disk_id}:/var/disk:rw")
            container_dst_uri = URL(f"/var/disk/{str(destination_url.path)[42:]}")
        else:
            container_dst_uri = destination_url

        async with get_neuro_client() as client:
            data_copier = DataCopier(client)
            job = await data_copier.launch(
                src_uri=container_src_uri,
                dst_uri=container_dst_uri,
                extract=extract,
                volume=volume,
                env=env,
                use_temp_dir=use_temp_dir,
                preset=preset,
                life_span=life_span,
            )
            exit_code = await _attach_job_stdout(job, client, name="copy")
            if exit_code == EX_OK:
                logger.info("Successfully copied data")
            else:
                raise click.ClickException(f"Data copy failed: {exit_code}")

    else:
        # otherwise we deal with cloud/local src and destination
        origin_src_url = source_url
        dst_dir = Path(destination_url.path).parent
        dst_name = origin_dst_name = Path(destination_url.path).name

        if use_temp_dir:
            # substitute original dst with TMP dst.
            # do extraction / compression there
            TEMP_UNPACK_DIR.mkdir(parents=True, exist_ok=True)
            dst_dir = Path(tempfile.mkdtemp(dir=str(TEMP_UNPACK_DIR)))
        if extract:
            # download data into subfolder of target for extraction
            dst_name = origin_dst_name + "/" + origin_src_url.name
        if compress:
            # preserve origin destination name for compressor
            if origin_dst_name == origin_src_url.name:
                logging.warning(
                    "Source file already has required archive extension. "
                    "Skipping compression step."
                )
                compress = False
            else:
                dst_name = origin_src_url.name

        cp_destination_url = URL.build(path=str(dst_dir / dst_name))
        await _nonstorage_cp(origin_src_url, cp_destination_url, remove_source=False)
        source_url = cp_destination_url

        if extract:
            dir_util.mkpath(str(dst_dir))
            extraction_dst_url = URL.build(path=str(dst_dir / origin_dst_name))
            await _extract(source_url, extraction_dst_url, rm_src=True)
            source_url = extraction_dst_url
        if compress:
            compression_dst_url = URL.build(path=str(dst_dir / origin_dst_name))
            await _compress(source_url, compression_dst_url, rm_src=True)
            source_url = compression_dst_url

        if use_temp_dir:
            # Move downloaded and maybe extracted / compressed files to
            # original destination.
            # Otherwise (if tmp was not used) - they are already there.
            await _nonstorage_cp(source_url, destination_url, remove_source=True)


def _patch_azure_url_for_rclone(url: URL) -> str:
    if url.scheme == "azure+https":
        return f":azureblob:{url.path}"
    else:
        return str(url)


def _build_sas_url(source_url: URL, destination_url: URL) -> URL:
    """
    In order to build SAS URL we replace original URL scheme with HTTPS,
    remove everything from path except bucket name and append SAS token as a query
    """
    azure_url = source_url if source_url.scheme == "azure+https" else destination_url
    sas_token = os.getenv("AZURE_SAS_TOKEN", "")
    azure_url = (
        azure_url.with_scheme("https")
        .with_path("/".join(azure_url.path.split("/")[:2]))
        .with_query(sas_token)
    )
    # with_query performs urlencode of sas_token, which brokes the token,
    # so we urldecode the resulting url
    azure_url = URL(parse.unquote(str(azure_url)))
    logger.info("Azure URL: %s", azure_url)
    return azure_url


async def _nonstorage_cp(
    source_url: URL, destination_url: URL, remove_source: bool = False
) -> None:
    if "s3" in (source_url.scheme, destination_url.scheme):
        command = "aws"
        args = ["s3", "cp", str(source_url), str(destination_url)]
        if source_url.path.endswith("/"):
            args.insert(2, "--recursive")
    elif "gs" in (source_url.scheme, destination_url.scheme):
        command = "gsutil"
        # gsutil service credentials are activated in entrypoint.sh
        args = ["-m", "cp", "-r", str(source_url), str(destination_url)]
    elif "azure+https" in (source_url.scheme, destination_url.scheme):
        sas_url = _build_sas_url(source_url, destination_url)
        command = "rclone"
        args = [
            "copyto",
            "--azureblob-sas-url",
            str(sas_url),
            _patch_azure_url_for_rclone(source_url),
            _patch_azure_url_for_rclone(destination_url),
        ]
    elif source_url.scheme in ("http", "https"):
        command = "rclone"
        args = [
            "copyto",
            "--http-url",
            # HTTP URL parameter for rclone is just scheme + host name
            str(source_url.with_path("").with_query("")),
            f":http:{source_url.path}",
            str(destination_url),
        ]
    elif source_url.scheme == "" and destination_url.scheme == "":
        command = "rclone"
        args = [
            "copyto",  # TODO: investigate usage of 'sync' for potential speedup.
            "--checkers=16",  # https://rclone.org/docs/#checkers-n , default is 8
            "--transfers=8",  # https://rclone.org/docs/#transfers-n , default is 4.
            "--verbose=1",  # default is 0, set 2 for debug
            str(source_url),
            str(destination_url),
        ]
    else:
        raise ValueError("Unknown cloud provider")
    click.echo(f"Running '{command} {' '.join(args)}'")
    subprocess = await asyncio.create_subprocess_exec(command, *args)
    returncode = await subprocess.wait()
    if UrlType.get_type(source_url) == UrlType.LOCAL and remove_source:
        _rm_local(Path(source_url.path))
    if returncode != 0:
        raise click.ClickException("Cloud copy failed")


async def _extract(source_url: URL, destination_url: URL, rm_src: bool) -> None:
    file = Path(source_url.path)
    if file.is_dir():
        file = list(file.glob("*"))[0]
    suffixes = file.suffixes
    if suffixes[-2:] == [".tar", ".gz"] or suffixes[-1] == ".tgz":
        command = "tar"
        args = ["zxvf", str(file), "-C", str(destination_url.path)]
    elif suffixes[-2:] == [".tar", ".bz2"] or suffixes[-1] in (".tbz2", ".tbz"):
        command = "tar"
        args = ["jxvf", str(file), "-C", str(destination_url)]
    elif suffixes[-1] == ".tar":
        command = "tar"
        args = ["xvf", str(file), "-C", str(destination_url)]
    elif suffixes[-1] == ".gz":
        command = "gunzip"
        args = ["--keep", str(file), str(destination_url) + file.name[:-3]]
    elif suffixes[-1] == ".zip":
        command = "unzip"
        args = [str(file), "-d", str(destination_url)]
    else:
        raise ValueError(
            f"Don't know how to extract file {file.name}"
            f"Supported archive types: {', '.join(SUPPORTED_ARCHIVE_TYPES)}"
        )

    click.echo(f"Running '{command} {' '.join(args)}'")
    subprocess = await asyncio.create_subprocess_exec(command, *args)
    returncode = await subprocess.wait()
    if rm_src:
        _rm_local(Path(source_url.path))
    if returncode != 0:
        raise click.ClickException(f"Extraction failed: {subprocess.stderr}")


async def _compress(source_url: URL, destination_url: URL, rm_src: bool) -> None:
    file = Path(destination_url.path.split("/")[-1])
    suffixes = file.suffixes
    if suffixes[-2:] == [".tar", ".gz"] or suffixes[-1] == ".tgz":
        command = "tar"
        args = [
            "zcf",
            str(destination_url.path),
            "-C",
            str(Path(source_url.path).parent),
            f"--exclude={str(destination_url.name)}",
            ".",
        ]
    elif suffixes[-2:] == [".tar", ".bz2"] or suffixes[-1] in (".tbz2", ".tbz"):
        command = "tar"
        args = [
            "jcf",
            str(destination_url.path),
            f"--exclude={str(destination_url.name)}",
            str(source_url.path),
        ]
    elif suffixes[-1] == ".tar":
        command = "tar"
        args = [
            "cf",
            str(destination_url.path),
            f"--exclude={str(destination_url.name)}",
            str(source_url.path),
        ]
    elif suffixes[-1] == ".gz":
        command = "gzip"
        args = ["-r", str(destination_url.path), str(source_url.path)]
    elif suffixes[-1] == ".zip":
        command = "zip"
        args = ["-r", str(destination_url.path), str(source_url.path)]
    else:
        raise ValueError(
            f"Don't know how to compress to archive type {file.name}. "
            f"Supported archive types: {', '.join(SUPPORTED_ARCHIVE_TYPES)}"
        )

    click.echo(f"Running '{command} {' '.join(args)}'")
    subprocess = await asyncio.create_subprocess_exec(command, *args)
    returncode = await subprocess.wait()
    if rm_src:
        _rm_local(Path(source_url.path))
    if returncode != 0:
        raise click.ClickException(f"Compression failed: {subprocess.stderr}")


def _rm_local(target: Path) -> None:
    # maybe also AWS / GCS clouds or storage?
    if target.is_dir():
        dir_util.remove_tree(str(target))
    if target.is_file() and target.exists():
        target.unlink()


async def _data_transfer(src_uri_str: str, dst_uri_str: str) -> None:
    src_cluster_or_null = _get_cluster_from_uri(src_uri_str, scheme="storage")
    dst_cluster = _get_cluster_from_uri(dst_uri_str, scheme="storage")

    if not src_cluster_or_null:
        async with get_neuro_client() as src_client:
            src_cluster = src_client.cluster_name
    else:
        src_cluster = src_cluster_or_null

    if not dst_cluster:
        raise ValueError(
            f"Invalid destination path {dst_uri_str}: missing cluster name"
        )

    async with get_neuro_client(cluster=dst_cluster) as client:
        await client.storage.mkdir(URL("storage:"), parents=True, exist_ok=True)
        await _run_copy_container(src_cluster, src_uri_str, dst_uri_str)


async def _run_copy_container(
    src_cluster: str, src_uri_str: str, dst_uri_str: str
) -> None:
    args = [
        "neuro",
        "run",
        "-s",
        "cpu-small",
        "--pass-config",
        "-v",
        f"{dst_uri_str}:/storage:rw",
        "-e",
        f"NEURO_CLUSTER={src_cluster}",  # inside the job, switch neuro to 'src_cluster'
        "--life-span 10d",
        NEURO_EXTRAS_IMAGE,
        "--",
        f"neuro --show-traceback cp --progress -r -u -T {src_uri_str} /storage",
    ]
    cmd = " ".join(args)
    click.echo(f"Running '{cmd}'")
    subprocess = await asyncio.create_subprocess_shell(cmd)
    returncode = await subprocess.wait()
    if returncode != 0:
        raise click.ClickException("Unable to copy storage")



from neuro_sdk import PluginManager

from .main import main  # noqa
from .version import __version__  # noqa


NEURO_EXTRAS_UPGRADE = """\
You are using Neuro Extras tool {old_ver}, however {new_ver} is available.
You should consider upgrading via the following command:
    python -m pip install --upgrade neuro-extras
"""


def get_neuro_extras_txt(old: str, new: str) -> str:
    return NEURO_EXTRAS_UPGRADE.format(old_ver=old, new_ver=new)


def setup_plugin(manager: PluginManager) -> None:
    manager.config.define_str("extra", "remote-project-dir")

    manager.version_checker.register("neuro-extras", get_neuro_extras_txt)



import asyncio
import logging
import os

import click
import neuro_sdk as neuro_api
from neuro_cli.const import EX_OK, EX_PLATFORMERROR

from .version import __version__


logger = logging.getLogger(__name__)

NEURO_EXTRAS_IMAGE = os.environ.get(
    "NEURO_EXTRAS_IMAGE", f"neuromation/neuro-extras:{__version__}"
)


async def _attach_job_stdout(
    job: neuro_api.JobDescription, client: neuro_api.Client, name: str = ""
) -> int:
    while job.status == neuro_api.JobStatus.PENDING:
        job = await client.jobs.status(job.id)
        await asyncio.sleep(1.0)
    async for chunk in client.jobs.monitor(job.id):
        if not chunk:
            break
        click.echo(chunk.decode(errors="ignore"), nl=False)
    while job.status in (neuro_api.JobStatus.PENDING, neuro_api.JobStatus.RUNNING):
        job = await client.jobs.status(job.id)
        await asyncio.sleep(1.0)

    job = await client.jobs.status(job.id)
    exit_code = EX_PLATFORMERROR
    if job.status == neuro_api.JobStatus.SUCCEEDED:
        exit_code = EX_OK
    elif job.status == neuro_api.JobStatus.FAILED:
        logger.error(f"The {name} job {job.id} failed due to:")
        logger.error(f"  Reason: {job.history.reason}")
        logger.error(f"  Description: {job.history.description}")
        exit_code = job.history.exit_code or EX_PLATFORMERROR  # never 0 for failed
    elif job.status == neuro_api.JobStatus.CANCELLED:
        logger.error(f"The {name} job {job.id} was cancelled")
    else:
        logger.error(f"The {name} job {job.id} terminated, status: {job.status}")
    return exit_code



import base64
import json
from pathlib import Path
from typing import Any, Dict

import click
import yaml
from neuro_cli.asyncio_utils import run as run_async

from .cli import main
from .image_builder import ImageBuilder
from .utils import get_neuro_client


@main.group()
def k8s() -> None:
    """
    Cluster Kubernetes operations.
    """
    pass


@k8s.command("generate-secret")
@click.option("--name", default="neuro")
def generate_k8s_secret(name: str) -> None:
    payload = run_async(_create_k8s_secret(name))
    click.echo(yaml.dump(payload), nl=False)


@k8s.command("generate-registry-secret")
@click.option("--name", default="neuro-registry")
def generate_k8s_registry_secret(name: str) -> None:
    payload = run_async(_create_k8s_registry_secret(name))
    click.echo(yaml.dump(payload), nl=False)


async def _create_k8s_secret(name: str) -> Dict[str, Any]:
    async with get_neuro_client() as client:
        payload: Dict[str, Any] = {
            "apiVersion": "v1",
            "kind": "Secret",
            "metadata": {"name": name},
            "type": "Opaque",
            "data": {},
        }
        config_path = Path(client.config._path)
        for path in config_path.iterdir():
            if path.is_dir() or path.name in ("db-shm", "db-wal"):
                continue
            payload["data"][path.name] = base64.b64encode(path.read_bytes()).decode()
        return payload


async def _create_k8s_registry_secret(name: str) -> Dict[str, Any]:
    async with get_neuro_client() as client:
        builder = ImageBuilder(client)
        docker_config = await builder.create_docker_config()
        return {
            "apiVersion": "v1",
            "kind": "Secret",
            "metadata": {"name": name},
            "type": "kubernetes.io/dockerconfigjson",
            "data": {
                ".dockerconfigjson": base64.b64encode(
                    json.dumps(docker_config.to_primitive()).encode()
                ).decode(),
            },
        }



import asyncio
import base64
import json
import logging
import re
import shlex
import uuid
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, AsyncIterator, Dict, Optional, Sequence, Tuple

import click
import neuro_sdk as neuro_api
from neuro_sdk.url_utils import normalize_storage_path_uri
from yarl import URL


KANIKO_IMAGE_REF = "gcr.io/kaniko-project/executor"
KANIKO_IMAGE_TAG = "v1.3.0-debug"  # since it has busybox, which is needed for auth
KANIKO_AUTH_PREFIX = "NE_REGISTRY_AUTH"
KANIKO_DOCKER_CONFIG_PATH = "/kaniko/.docker/config.json"
KANIKO_AUTH_SCRIPT_PATH = "/kaniko/.docker/merge_docker_auths.sh"
KANIKO_CONTEXT_PATH = "/kaniko_context"
BUILDER_JOB_LIFESPAN = "4h"
BUILDER_JOB_SHEDULE_TIMEOUT = "20m"

logger = logging.getLogger(__name__)


@dataclass
class DockerConfigAuth:
    registry: str
    username: str
    password: str = field(repr=False)

    @property
    def credentials(self) -> str:
        return base64.b64encode(f"{self.username}:{self.password}".encode()).decode()


@dataclass
class DockerConfig:
    auths: Sequence[DockerConfigAuth] = ()

    def to_primitive(self) -> Dict[str, Any]:
        return {
            "auths": {auth.registry: {"auth": auth.credentials} for auth in self.auths}
        }


async def create_docker_config_auth(
    client_config: neuro_api.Config,
) -> DockerConfigAuth:
    # retrieve registry hostname with optional port
    url = client_config.registry_url
    assert url.host
    port = f":{url.explicit_port}" if url.explicit_port else ""  # type: ignore
    registry_host = url.host + port
    auth = DockerConfigAuth(
        registry=registry_host,
        username=client_config.username,
        password=await client_config.token(),
    )
    return auth


class ImageBuilder:
    def __init__(
        self,
        client: neuro_api.Client,
        extra_registry_auths: Sequence[DockerConfigAuth] = (),
        verbose: bool = False,
    ) -> None:
        """
            Builds and pushes docker image to the platform.
            Build is currently happens on the platform, using Kaniko tool.
            TODO: add possibility to build using local docker.

        Args:
            client (neuro_api.Client): instance of neuro-sdk client,
                authenticated to the destination cluster
            extra_registry_auths (Sequence[DockerConfigAuth], optional):
                Sequence of extra docker container registry auth credits,
                useful if base image(s) hidden under private registry(es).
                Defaults to ().
            verbose (bool, optional): Whether to set Kaniko's verbosity to DEBUG.
                Defaults to False.
        """
        self._client = client
        self._extra_registry_auths = list(extra_registry_auths)
        self._verbose = verbose

    def _generate_build_uri(self) -> URL:
        return normalize_storage_path_uri(
            URL(f"storage:.builds/{uuid.uuid4()}"),
            self._client.username,
            self._client.cluster_name,
        )

    async def create_docker_config(self) -> DockerConfig:
        dst_reg_auth = await create_docker_config_auth(self._client.config)
        return DockerConfig(auths=[dst_reg_auth] + self._extra_registry_auths)

    async def save_docker_config(self, docker_config: DockerConfig, uri: URL) -> None:
        async def _gen() -> AsyncIterator[bytes]:
            yield json.dumps(docker_config.to_primitive()).encode()

        await self._client.storage.create(uri, _gen())

    def parse_image_ref(self, image_uri_str: str) -> str:
        image = self._client.parse.remote_image(image_uri_str)
        return re.sub(r"^http[s]?://", "", image.as_docker_url())

    async def build(
        self,
        dockerfile_path: Path,
        context_uri: URL,
        image_uri_str: str,
        use_cache: bool,
        build_args: Tuple[str, ...],
        volumes: Tuple[str, ...],
        envs: Tuple[str, ...],
        job_preset: Optional[str],
        build_tags: Tuple[str, ...],
    ) -> int:
        # TODO: check if Dockerfile exists
        logger.info(f"Building the image {image_uri_str}")
        logger.info(f"Using {context_uri} as the build context")

        # upload (if needed) build context and platform registry auth info
        build_uri = self._generate_build_uri()
        await self._client.storage.mkdir(build_uri, parents=True)
        if context_uri.scheme == "file":
            local_context_uri, context_uri = context_uri, build_uri / "context"
            logger.info(f"Uploading {local_context_uri} to {context_uri}")
            subprocess = await asyncio.create_subprocess_exec(
                "neuro",
                "--disable-pypi-version-check",
                "cp",
                "--recursive",
                str(local_context_uri),
                str(context_uri),
            )
            return_code = await subprocess.wait()
            if return_code != 0:
                raise click.ClickException("Uploading build context failed!")

        docker_config = await self.create_docker_config()
        docker_config_uri = build_uri / ".docker.config.json"
        logger.debug(f"Uploading {docker_config_uri}")
        await self.save_docker_config(docker_config, docker_config_uri)

        cache_image = neuro_api.RemoteImage(
            name="layer-cache/cache",
            owner=self._client.config.username,
            registry=str(self._client.config.registry_url),
            cluster_name=self._client.cluster_name,
        )
        cache_repo = self.parse_image_ref(str(cache_image))
        cache_repo = re.sub(r":.*$", "", cache_repo)  # drop tag

        if any(KANIKO_AUTH_PREFIX in env for env in envs):
            # we have extra auth info.
            # in this case we cannot mount registry auth info at the default path
            # and should upload and configure 'merge_docker_auths' script to merge auths
            mnt_path = Path(KANIKO_DOCKER_CONFIG_PATH)
            mnt_path = mnt_path.with_name(f"{mnt_path.stem}_base{mnt_path.suffix}")
            docker_config_mnt = str(mnt_path)
            envs += (
                f"{KANIKO_AUTH_PREFIX}_BASE_{uuid.uuid4().hex[:8]}={docker_config_mnt}",
            )
            local_script = URL(
                (Path(__file__).parent / "assets" / "merge_docker_auths.sh").as_uri()
            )
            remote_script = build_uri / "merge_docker_auths.sh"
            await self._client.storage.upload_file(local_script, remote_script)
            volumes += (f"{remote_script}:{KANIKO_AUTH_SCRIPT_PATH}:ro",)
            entrypoint = [
                f"sh {KANIKO_AUTH_SCRIPT_PATH}",
                "&&",
                "executor",
                # Kaniko args will be added below
            ]
        else:
            docker_config_mnt = str(KANIKO_DOCKER_CONFIG_PATH)
            entrypoint = []

        # mount build context and Kaniko auth info
        volumes += (
            f"{docker_config_uri}:{docker_config_mnt}:ro",
            # context dir cannot be R/O if we want to mount secrets there
            f"{context_uri}:{KANIKO_CONTEXT_PATH}:rw",
        )
        dst_image = self._client.parse.remote_image(image_uri_str)
        build_tags += (f"kaniko-builds-image:{dst_image}",)
        kaniko_args = [
            f"--dockerfile={KANIKO_CONTEXT_PATH}/{dockerfile_path}",
            f"--destination={self.parse_image_ref(image_uri_str)}",
            f"--cache={'true' if use_cache else 'false'}",
            # f"--cache-copy-layers", # TODO: since kaniko 1.3 does not support it
            f"--cache-repo={cache_repo}",
            f"--snapshotMode=redo",
            f"--verbosity={'debug' if self._verbose else 'info'}",
            f"--context={KANIKO_CONTEXT_PATH}",
        ]

        for arg in build_args:
            kaniko_args.append(f"--build-arg {arg}")
        # env vars (which might be platform secrets too) are passed as build args
        env_parsed = self._client.parse.envs(envs)
        for arg in list(env_parsed.env) + list(env_parsed.secret_env):
            if KANIKO_AUTH_PREFIX not in arg:
                kaniko_args.append(f"--build-arg {arg}")

        build_command = [
            "neuro",
            "--disable-pypi-version-check",
            "job",
            "run",
            f"--life-span={BUILDER_JOB_LIFESPAN}",
            f"--schedule-timeout={BUILDER_JOB_SHEDULE_TIMEOUT}",
        ]
        if job_preset:
            build_command.append(f"--preset={job_preset}")
        for volume in volumes:
            build_command.append(f"--volume={volume}")
        for env in envs:
            build_command.append(f"--env={env}")
        for build_tag in build_tags:
            build_command.append(f"--tag={build_tag}")
        if entrypoint:
            entrypoint.append(" ".join(kaniko_args))
            build_command.append("--entrypoint")
            build_command.append(f"sh -c {shlex.quote(' '.join(entrypoint))}")
            build_command.append(f"{KANIKO_IMAGE_REF}:{KANIKO_IMAGE_TAG}")
        else:
            build_command.append(f"{KANIKO_IMAGE_REF}:{KANIKO_IMAGE_TAG}")
            build_command.append("--")
            build_command.append(" ".join(kaniko_args))

        logger.info("Submitting a builder job")
        logger.debug(build_command)
        subprocess = await asyncio.create_subprocess_exec(*build_command)
        # TODO: remove context after the build is finished?
        return await subprocess.wait()



from distutils import dir_util
from pathlib import Path
from typing import Any, Dict

import click
import yaml
from neuro_cli.asyncio_utils import run as run_async
from neuro_sdk.url_utils import uri_from_cli
from yarl import URL

from .cli import main
from .common import NEURO_EXTRAS_IMAGE
from .image_builder import ImageBuilder
from .utils import get_neuro_client


ASSETS_PATH = Path(__file__).resolve().parent / "assets"
SELDON_CUSTOM_PATH = ASSETS_PATH / "seldon.package"


@main.group()
def seldon() -> None:
    """
    Seldon deployment operations.
    """
    pass


@seldon.command("init-package")
@click.argument("path", default=".")
def seldon_init_package(path: str) -> None:
    run_async(_init_seldon_package(path))


@seldon.command("generate-deployment")
@click.option("--name", default="neuro-model")
@click.option("--neuro-secret", default="neuro")
@click.option("--registry-secret", default="neuro-registry")
@click.argument("model-image-uri")
@click.argument("model-storage-uri")
def generate_seldon_deployment(
    name: str,
    neuro_secret: str,
    registry_secret: str,
    model_image_uri: str,
    model_storage_uri: str,
) -> None:
    payload = run_async(
        _create_seldon_deployment(
            name=name,
            neuro_secret_name=neuro_secret,
            registry_secret_name=registry_secret,
            model_image_uri=model_image_uri,
            model_storage_uri=model_storage_uri,
        )
    )
    click.echo(yaml.dump(payload), nl=False)


async def _init_seldon_package(path: str) -> None:
    async with get_neuro_client() as client:
        uri = uri_from_cli(
            path,
            client.username,
            client.cluster_name,
            allowed_schemes=("file", "storage"),
        )
        click.echo(f"Copying a Seldon package scaffolding into {uri}")
        if uri.scheme == "file":
            dir_util.copy_tree(str(SELDON_CUSTOM_PATH), path)
        else:
            await client.storage.mkdir(uri, parents=True)
            await client.storage.upload_dir(URL(SELDON_CUSTOM_PATH.as_uri()), uri)


async def _create_seldon_deployment(
    *,
    name: str,
    neuro_secret_name: str,
    registry_secret_name: str,
    model_image_uri: str,
    model_storage_uri: str,
) -> Dict[str, Any]:
    async with get_neuro_client() as client:
        builder = ImageBuilder(client)
        model_image_ref = builder.parse_image_ref(model_image_uri)

    pod_spec = {
        "volumes": [
            {"emptyDir": {}, "name": "neuro-storage"},
            {"name": "neuro-secret", "secret": {"secretName": neuro_secret_name}},
        ],
        "imagePullSecrets": [{"name": registry_secret_name}],
        "initContainers": [
            {
                "name": "neuro-download",
                "image": NEURO_EXTRAS_IMAGE,
                "imagePullPolicy": "Always",
                "securityContext": {"runAsUser": 0},
                "command": ["bash", "-c"],
                "args": [
                    "cp -L -r /var/run/neuro/config /root/.neuro;"
                    "chmod 0700 /root/.neuro;"
                    "chmod 0600 /root/.neuro/db;"
                    f"neuro cp {model_storage_uri} /storage"
                ],
                "volumeMounts": [
                    {"mountPath": "/storage", "name": "neuro-storage"},
                    {"mountPath": "/var/run/neuro/config", "name": "neuro-secret"},
                ],
            }
        ],
        "containers": [
            {
                "name": "model",
                "image": model_image_ref,
                "imagePullPolicy": "Always",
                "volumeMounts": [{"mountPath": "/storage", "name": "neuro-storage"}],
            }
        ],
    }
    return {
        "apiVersion": "machinelearning.seldon.io/v1",
        "kind": "SeldonDeployment",
        "metadata": {"name": name},
        "spec": {
            "predictors": [
                {
                    "componentSpecs": [{"spec": pod_spec}],
                    "graph": {
                        "endpoint": {"type": "REST"},
                        "name": "model",
                        "type": "MODEL",
                    },
                    "name": "predictor",
                    "replicas": 1,
                }
            ]
        },
    }



import re

from setuptools import find_packages, setup


with open("neuro_extras/version.py") as f:
    txt = f.read()
    try:
        version = re.findall(r'^__version__ = "([^"]+)"\r?$', txt, re.M)[0]
    except IndexError:
        raise RuntimeError("Unable to determine version.")


setup(
    name="neuro-extras",
    version=version,
    python_requires=">=3.6.0",
    url="https://github.com/neuro-inc/neuro-extras",
    packages=find_packages(),
    install_requires=[
        "neuro-cli>=21.11.2",
        'dataclasses>=0.7; python_version<"3.7"',
        "click>=7.0",
        "toml>=0.10.0",
        "pyyaml>=3.0",
    ],
    entry_points={
        "console_scripts": ["neuro-extras=neuro_extras:main"],
        "neuro_api": ["neuro-extras=neuro_extras:setup_plugin"],
    },
    zip_safe=False,
    include_package_data=True,
)



import logging
from pathlib import Path
from typing import Any, MutableMapping

import toml

from .cli import main
from .config import save_registry_auth  # noqa
from .data import data_cp, data_transfer  # noqa
from .image import image_build, image_transfer  # noqa
from .k8s import generate_k8s_registry_secret, generate_k8s_secret  # noqa
from .seldon import generate_seldon_deployment, seldon_init_package  # noqa
from .upload_download import download, upload  # noqa


logger = logging.getLogger(__name__)


@main.command("init-aliases")
def init_aliases() -> None:
    """
    Create neuro CLI aliases for neuro-extras functionality.
    """
    # TODO: support patching the global ~/.neuro/user.toml
    toml_path = Path.cwd() / ".neuro.toml"
    config: MutableMapping[str, Any] = {}
    if toml_path.exists():
        with toml_path.open("r") as f:
            config = toml.load(f)
    config.setdefault("alias", {})
    config["alias"]["image-build"] = {
        "exec": "neuro-extras image build",
        "options": [
            "-f, --file=PATH  path to the Dockerfile within CONTEXT",
            "--build-arg=LIST  build arguments for Docker",
            "--build-tag=LIST  tag(s) for image builder job.",
            "-e, --env=LIST  environment variables for container",
            "-v, --volume=LIST  list of volumes for container",
            "-s, --preset=STR  specify  preset for builder container",
            "-F, --force-overwrite  enforce destination image overwrite",
        ],
        "args": "CONTEXT_PATH IMAGE_URI",
        "help": (
            "Build docker image on the platform. "
            "Hit `neuro-extras image build --help` for more info."
        ),
    }
    config["alias"]["seldon-init-package"] = {
        "exec": "neuro-extras seldon init-package",
        "args": "URI_OR_PATH",
    }
    config["alias"]["image-transfer"] = {
        "exec": "neuro-extras image transfer",
        "args": "SOURCE DESTINATION",
        "options": [
            "-F, --force-overwrite  enforce destination image overwrite",
        ],
        "help": (
            "Transfer images between the cluster within the platform. "
            "Hit `neuro-extras image transfer --help` for more info."
        ),
    }
    config["alias"]["data-transfer"] = {
        "exec": "neuro-extras data transfer",
        "args": "SOURCE DESTINATION",
    }
    config["alias"]["data-cp"] = {
        "exec": "neuro-extras data cp",
        "options": [
            "-c, --compress Compress source files",
            "-x, --extract Extract downloaded files",
            "-e, --env environment variables for container",
            "-v, --volume list of volumes for container",
            "-t, --use-temp-dir store intermediate data in TMP directory",
            "-s, --preset=NAME  specify  preset for copy container",
            "-l, --life_span=MINUTES  specify life span for copy container",
        ],
        "args": "SOURCE DESTINATION",
    }
    with toml_path.open("w") as f:
        toml.dump(config, f)
    logger.info(f"Added aliases to {toml_path}")
